"""Contains tests for the collections module.
"""
import datetime
import logging
import os
import sys
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries import collections
logger = logging.getLogger(__name__)


class TestDictList(AriesTest):
    """Contains tests for the DictList class
    """
    dict_list = [
        {
            "name": "A",
            "order": 1,
        },
        {
            "name": "B",
            "order": 2,
        },
        {
            "name": "X",
            "order": 0,
        },
    ]

    def test_sort_dict_list(self):
        """Tests sorting a list of dictionaries.
        """
        dl = collections.DictList(self.dict_list)
        dl = dl.sort_by_value("order")
        self.assertEqual(dl[0], {"name": "X", "order": 0})
        self.assertEqual(dl[2], {"name": "B", "order": 2})

    def test_get_unique_keys(self):
        """Test getting unique keys from a list of dictionaries.
        """
        dl = collections.DictList(self.dict_list)
        keys = dl.unique_keys()
        self.assertIn("order", keys)
        self.assertIn("name", keys)
        self.assertEqual(len(keys), 2)


class TestLists(AriesTest):
    def test_sort_lists(self):
        """Tests sorting two lists
        """
        order_list = [2, 1, 3]
        label_list = ['A', 'B', 'C']
        order_list, label_list = collections.sort_lists(order_list, label_list)
        self.assertEqual(order_list, [1, 2, 3])
        self.assertEqual(label_list, ['B', 'A', 'C'])



"""Contains tests for the web module.
"""
import datetime
import logging
import os
import sys
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries import web
logger = logging.getLogger(__name__)


class TestWeb(AriesTest):
    def test_web_api_get(self):
        api = web.WebAPI("https://api.weather.gov/")
        res = api.get_json("")
        self.assertEqual(res.get("status"), "OK")

    def test_download(self):
        file_path = os.path.join(os.path.dirname(__file__), "test_download")
        if os.path.exists(file_path):
            os.remove(file_path)
        web.download("https://www.google.com", file_path)
        self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)

    def test_get_html_table(self):
        url = "https://en.wikipedia.org/wiki/List_of_file_signatures"
        tables = web.HTML(url).get_tables()
        self.assertEqual(len(tables), 2, "There should be two tables in the HTML page.")



import requests
import re
import logging
from .web import WebAPI
logger = logging.getLogger(__name__)


class DockerAPI(WebAPI):
    def __init__(self, hostname):
        super().__init__("https://%s" % hostname)

    def check_version(self, endpoint="/v2/"):
        """Checks the API version by sending GET request to /v2/ endpoint.

        Returns: The API version number. e.g. 2.0

        """
        response = self.get(endpoint)
        version = response.headers.get("Docker-Distribution-Api-Version", "").split("/")[-1]
        return version

    @staticmethod
    def get_auth_url(authenticate_header):
        """Gets the authentication URL from the www-authenticate header value.

        Args:
            authenticate_header:

        Returns:

        """
        authenticate_header = str(authenticate_header)
        # Only Bearer token authentication is supported.
        if not authenticate_header.startswith("Bearer"):
            raise NotImplementedError("Authentication method not supported: %s" % authenticate_header)
        value = authenticate_header[len("Bearer"):].strip()
        # Convert the values to a dictionary
        auth_dict = dict()
        for v in value.split(","):
            if "=" not in v:
                continue
            v = v.strip()
            arr = v.split('=', 1)
            key = arr[0]
            val = arr[1].strip('"')
            values = auth_dict.get(key, [])
            values.append(val)
            auth_dict[key] = values
        # logger.debug(auth_dict)
        realm = auth_dict.get("realm")
        if not realm:
            raise ValueError("realm not found in www-authenticate header.")
        endpoint = realm[0]
        params = dict()
        for key in ["scope", "service"]:
            if key in auth_dict:
                params[key] = auth_dict[key]
        return WebAPI.append_query_string(endpoint, **params)

    def request(self, method, url, **kwargs):
        """
        This method uses the same arguments as the python requests package
        https://github.com/psf/requests/blob/master/requests/api.py

        Args:
            method: Request method, e.g. GET, OPTIONS, HEAD, POST, PUT, PATCH, or DELETE
            url: URL endpoint for the request, which can be relative URL.
            **kwargs: See https://github.com/psf/requests/blob/master/requests/api.py

        Returns: Request response

        """
        response = super().request(method, url, **kwargs)
        # Try to get authentication token automatically for public repository
        # TODO: Authentication for private repository
        if response.status_code == 401:
            auth_url = self.get_auth_url(response.headers.get("www-authenticate"))
            logger.debug("Requesting token at: %s" % auth_url)
            auth_res = requests.get(auth_url).json()
            token = auth_res.get("token")
            # Adds the token to headers so that it will be used in the subsequent requests
            self.add_header(Authorization="Bearer %s" % token)
            response = super().request(method, url, **kwargs)
        return response


class DockerImage:
    """
    Attributes:
        name: The docker image name, which may contain the hostname if it is not on DockerHub,
            and optionally the tag or digest.
            This is the same as the NAME[:TAG|@DIGEST] to be used in the docker pull command.
            e.g. ubuntu, tensorflow/tensorflow:2.2.0rc4-jupyter, mcr.microsoft.com/windows
            See Also: https://docs.docker.com/engine/reference/commandline/pull/
        hostname: The hostname of the Docker registry.
        path: The path of the docker image under the registry, without tag or digest.
        tag: The docker image tag, if included in the name.

    See Also:
        https://docs.docker.com/registry/spec/api/
        https://github.com/docker/distribution/blob/master/reference/reference.go
        https://docs.docker.com/engine/reference/commandline/pull/
        https://ops.tips/blog/inspecting-docker-image-without-pull/
    """

    DOCKER_IO_REGISTRY = "registry-1.docker.io"

    def __init__(self, name):
        self.name = name
        self.hostname, self.path, self.tag, self._digest = self.parse_name(self.name)
        self.api = DockerAPI(self.hostname)
        self._manifest = None

    @staticmethod
    def parse_name(name):
        """Parses docker image name, which can have the format as accepted by the docker pull command.

        Args:
            name (str): Docker image name, including tag or digest, i.e. NAME[:TAG|@DIGEST]

        Returns:
            A 4-tuple of hostname, path, tag and digest.
            hostname defaults to DOCKER_IO_REGISTRY
        """
        hostname, path = DockerImage.match_host(name)
        if not hostname:
            hostname = DockerImage.DOCKER_IO_REGISTRY

        # Try to match the digest followed by the image name.
        path, digest = DockerImage.match_digest(path)

        # There should be no tag if digest is followed by the image name.
        if digest:
            tag = None
        else:
            path, tag = DockerImage.match_tag(path)

        # Official images are identified with library/IMAGE_NAME
        if "/" not in path and hostname == DockerImage.DOCKER_IO_REGISTRY:
            path = "library/" + path

        return hostname, path, tag, digest

    @staticmethod
    def match_host(name):
        """Parses the docker image name, extracts the hostname and path (which may also include the tag/digest).
        """
        name_pattern = r"(?:((?:localhost)|(?:[^/]*\.[^/]*)|(?:[^/]*\:[^/]*))/)?(.+)"
        match = re.match(name_pattern, name)
        if not match:
            raise ValueError("Invalid image name: %s" % name)
        groups = match.groups()
        hostname = groups[0]
        path_with_tag = groups[1]
        return hostname, path_with_tag

    @staticmethod
    def match_tag(name):
        tag_pattern = r"(.+)\:([\w][\w.-]{0,127})"
        match = re.match(tag_pattern, name)
        if match:
            path = match.groups()[0]
            tag = match.groups()[1]
        else:
            path = name
            tag = "latest"
        return path, tag

    @staticmethod
    def match_digest(name):
        digest_pattern = r"(.+)\@([^/]*)"
        match = re.match(digest_pattern, name)
        if match:
            path = match.groups()[0]
            digest = match.groups()[1]
        else:
            path = name
            digest = None
        return path, digest

    @property
    def digest(self):
        """The digest of the docker image
        """
        if not self._digest:
            self._digest = self.get_manifest().headers.get("Docker-Content-Digest").strip()
        return self._digest

    def get_manifest(self):
        """Gets the manifest HTTP response.
        """
        if not self._manifest:
            reference = "latest"
            if self.tag:
                reference = self.tag
            elif self.digest:
                reference = self.digest
            url = "/v2/%s/manifests/%s" % (self.path, reference)
            response = self.api.request(
                "GET",
                url,
                headers={"Accept": "application/vnd.docker.distribution.manifest.v2+json"}
            )
            if response.status_code != 200:
                raise ValueError(
                    "Request to %s failed with status code %s" % (url, response.status_code)
                )
            self._manifest = response
        return self._manifest

    def inspect(self):
        config_digest = self.get_manifest().json().get("config", {}).get("digest")
        url = "/v2/%s/blobs/%s" % (self.path, config_digest)
        return self.api.request("GET", url).json()

    def is_accessible(self):
        """Checks if the image is accessible.

        Returns: True if the image is accessible.
            This method returns False if the existence cannot be determined.

        """
        try:
            self.get_manifest()
            return True
        except ValueError:
            return False

    def get_size(self):
        """Gets the size of the docker image by adding the sizes of all layers.

        Returns: The size of the docker image in bytes.
            This method returns None if the size cannot be determined.
        """
        try:
            # layers is defined in Image Manifest Version 2, Schema 2
            # See also: https://docs.docker.com/registry/spec/manifest-v2-2/
            layers = self.get_manifest().json().get("layers")
            if not layers:
                return self.get_size_via_fs_layers()
            total_size = 0
            for layer in layers:
                total_size += layer.get("size", 0)
        except Exception as ex:
            logger.error("Cannot determine image size for %s: %s" % (self.name, str(ex)))
            import traceback
            traceback.print_exc()
            return None
        return total_size

    def get_size_via_fs_layers(self):
        """Gets the size of the docker image by adding the sizes of all fsLayers
        This is designed for Image Manifest Version 2, Schema 1

        Returns: The size of the docker image in bytes.
            This method returns None if the size cannot be determined.

        See Also: https://docs.docker.com/registry/spec/manifest-v2-1/

        """
        layers = self.get_manifest().json().get("fsLayers")
        if not layers:
            logger.error("Cannot determine image size for %s" % self.name)
            return None
        size = 0
        for layer in layers:
            digest = layer.get("blobSum")
            if not digest:
                continue
            res = self.api.request("HEAD", "/v2/%s/blobs/%s" % (self.path, digest))
            layer_size = res.headers.get("Content-Length")
            if layer_size and str(layer_size).isdigit():
                size += int(layer_size)
        return size



import logging
import os
import boto3
import sys
import traceback
import time
from botocore.exceptions import NoCredentialsError
try:
    from ..test import AriesTest
    from ..storage import StorageFolder, StorageFile
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.test import AriesTest
    from Aries.storage import StorageFolder, StorageFile
    from Aries.storage.gs import GSObject
    from Aries.strings import Base64String
logger = logging.getLogger(__name__)


class TestGCStorage(AriesTest):
    """Contains test cases for Google Cloud Platform Storage.
    """
    AWS_CREDENTIALS = None

    TEST_BUCKET_NAME = "davelab-test"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        client = boto3.client('s3')
        try:
            client.list_buckets()
            cls.AWS_CREDENTIALS = True
        except NoCredentialsError as ex:
            print("AWS Credentials not found.")
            print("%s: %s" % (type(ex), str(ex)))
            traceback.print_exc()

    def setUp(self):
        # Skip test if AWS credentials are not found.
        if not self.AWS_CREDENTIALS:
            self.skipTest("AWS Credentials not found.")

    def test_parse_uri(self):
        """Tests parsing GCS URI
        """
        # File
        file_obj = StorageFile("s3://%s/test_file.txt" % self.TEST_BUCKET_NAME)
        self.assertEqual(file_obj.scheme, "s3")
        self.assertEqual(file_obj.path, "/test_file.txt")

        # Folder
        folder_obj = StorageFolder("s3://%s/test_folder" % self.TEST_BUCKET_NAME)
        self.assertEqual(folder_obj.uri, "s3://%s/test_folder/" % self.TEST_BUCKET_NAME)
        self.assertEqual(folder_obj.scheme, "s3")
        self.assertEqual(folder_obj.path, "/test_folder/")

        # Bucket root
        folder_obj = StorageFolder("s3://%s" % self.TEST_BUCKET_NAME)
        self.assertEqual(folder_obj.uri, "s3://%s/" % self.TEST_BUCKET_NAME)
        self.assertEqual(folder_obj.scheme, "s3")
        self.assertEqual(folder_obj.path, "/")





"""Provides unified shortcuts/interfaces for access folders and files.
"""
import os
import logging
from abc import ABC
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse
from io import RawIOBase, UnsupportedOperation, SEEK_SET, DEFAULT_BUFFER_SIZE
logger = logging.getLogger(__name__)


class StorageObject:
    """Represents a storage object.
    This is the base class for storage folder and storage file.

    """
    # Use a large buffer to improve performance of cloud storage access.
    BUFFER_SIZE = DEFAULT_BUFFER_SIZE * 128

    def __init__(self, uri):
        """Initializes a storage object.

        Args:
            uri (str): Uniform Resource Identifier for the object.
            The uri should include a scheme, except local files.

        See https://en.wikipedia.org/wiki/Uniform_Resource_Identifier
        """
        self.uri = str(uri)
        parse_result = urlparse(self.uri)
        self.scheme = parse_result.scheme
        self.hostname = parse_result.hostname
        self.path = parse_result.path

        # Use file as scheme if one is not in the URI
        if not self.scheme:
            self.scheme = 'file'
            self.hostname = ""
            if not uri.startswith("/"):
                self.uri = os.path.abspath(self.uri)
            self.path = self.uri
            self.uri = "file://" + self.uri

        # The "prefix" for does not include the beginning "/"
        if self.path.startswith("/"):
            self.prefix = self.path[1:]
        else:
            self.prefix = self.path

    def __str__(self):
        """Returns the URI
        """
        return self.uri

    def __repr__(self):
        return self.uri

    @property
    def basename(self):
        """The basename of the file/folder, without path or "/".

        Returns:
            str: The basename of the file/folder
        """
        return os.path.basename(self.path.strip("/"))

    @property
    def name(self):
        """The basename of the file/folder, without path or "/".
        Same as basename.

        Returns:
            str: The basename of the file/folder
        """
        return self.basename

    @staticmethod
    def get_attributes(storage_objects, attribute):
        """Gets the attributes of a list of storage objects.

        Args:
            storage_objects (list): A list of Storage Objects, from which the values of an attribute will be extracted.
            attribute (str): A attribute of the storage object.

        Returns (list): A list of attribute values.

        """
        if not storage_objects:
            return []
        elif not attribute:
            return [str(f) for f in storage_objects]
        else:
            return [getattr(f, attribute) for f in storage_objects]

    @staticmethod
    def copy_stream(from_file_obj, to_file_obj):
        """Copies data from one file object to another
        """
        # TODO: there could be a problem if the file_obj has a different buffer size.
        chunk_size = StorageObject.BUFFER_SIZE
        file_size = 0
        while True:
            b = from_file_obj.read(chunk_size)
            if not b:
                break
            file_size += to_file_obj.write(b)
        to_file_obj.flush()
        return file_size

    def create_temp_file(self, delete=False, **kwargs):
        """Creates a NamedTemporaryFile on local computer with the same file extension.
        Everything after the first dot is considered as extension
        """
        # Determine the file extension
        if "suffix" not in kwargs:
            arr = self.basename.split(".", 1)
            if len(arr) > 1:
                suffix = ".%s" % arr[1]
                kwargs["suffix"] = suffix

        temp_obj = NamedTemporaryFile('w+b', delete=delete, **kwargs)
        logger.debug("Created temp file: %s" % temp_obj.name)
        return temp_obj


class StoragePrefixBase(StorageObject):
    @property
    def objects(self):
        """All storage files under this object, e.g. all files in the folder and sub-folders

        Returns: A list of storage file objects.

        """
        from .io import StorageFile
        return [StorageFile(uri) for uri in self.uri_list]

    @property
    def uri_list(self):
        """A list of URIs for each object with the same prefix
        """
        raise NotImplementedError()

    @property
    def size(self):
        """The size in bytes of all objects with the same prefix.

        Returns:

        This method requires the "object" to have a size attribute,
            sub-class should overwrite this if the size attribute is not available.

        AWS CLI: aws s3 ls --summarize --human-readable --recursive s3://bucket_name/path/to/folder | grep 'Total'
        GCP CLI: gsutil du -s gs://bucket_name/path/to/folder

        """
        total = 0
        for obj in self.objects:
            s = obj.size
            total += s if s else 0
        return total

    def exists(self):
        return True if self.objects else False

    def delete(self):
        for obj in self.objects:
            if not hasattr(obj, "delete") or not callable(obj.delete):
                raise UnsupportedOperation("Object %s does not support delete() method." % obj)
            obj.delete()


class StorageFolderBase(StorageObject):
    """Represents a folder/container of storage objects.
    This is the base class for folders or folder-like objects.
    It is used by
        Raw platform dependent classes, e.g. GSFolder, S3Folder, and
        IO interface class, i.e. StorageFolder
    When implementing raw, platform dependent sub-class:
        All the abstract methods should be implemented, otherwise a NotImplementedError will be raised:
            file_paths
            folder_paths
            exists()
            create()
            delete()
        Other methods (raising UnsupportedOperation error) can be implemented optionally to improve the performance:
            objects()
            copy()

    The io.StorageFolder class contains default implementations for methods like objects() and copy().
    In io.StorageFolder, when these methods are not implemented in the raw class,
        the UnsupportedOperation error will be caught and the default implementations will be used.
    """
    def __init__(self, uri):
        # Make sure uri ends with "/" for folders
        if uri and uri[-1] != '/':
            uri += '/'
        StorageObject.__init__(self, uri)

    @property
    def file_paths(self):
        """

        Returns: A list of URIs, each points to a file in the folder.

        """
        raise NotImplementedError()

    @property
    def folder_paths(self):
        """

        Returns: A list of URIs, each points to a folder in the folder.

        """
        raise NotImplementedError()

    def exists(self):
        """Checks if the folder exists.
        """
        raise NotImplementedError()

    def create(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    @property
    def objects(self):
        raise UnsupportedOperation()

    def copy(self, to, contents_only=False):
        """Copies the folder to another location within the same scheme
        """
        raise UnsupportedOperation()


class StorageIOBase(StorageObject, RawIOBase):
    """Base class designed to provide:
        1. The underlying RawIO for a BufferedIO.
        2. High level operations like copy() and delete().

    StorageIOBase is an extension of the python RawIOBase
    See Also: https://docs.python.org/3/library/io.html#class-hierarchy

    The RawIO in StorageIOBase implementation is similar to the implementation of FileIO
    A sub-class of StorageIOBase can be used in place of FileIO
    Each sub-class should implement:
        read(), for reading bytes from the file.
        write(), for writing bytes into the file.
        close(), for closing the file.
        open(), should also be implemented if needed.

    In addition to interface provided by RawIOBase,
    StorageIOBase also defines some high level APIs.
    For high level operations, a sub-class should implement:
        size, the size of the file in bytes .
        exists(), determine if a file exists.
        delete(), to delete the file.
        load_from(), to load/create the file from a stream.

    Optionally, the following methods can be implemented
        to speed up the corresponding high-level operations.
        copy(), if a sub-class implements copy(), it should create the destination folder automatically when needed.
        local()
        upload()
        download()

    StorageIOBase and its sub-classes are intended to be the underlying raw IO of StorageFile.
    In general, they should not be used directly. The StorageFile class should be used instead.

    The file is NOT opened when initializing StorageIOBase with __init__().
    To open the file, call open() or use StorageIOBase.init().
    The close() method should be called after writing data into the file.
    Alternatively, the context manager can be used, e.g. "with StorageIOBase(uri) as f:"


    See Also:
        https://docs.python.org/3/library/io.html#io.FileIO
        https://docs.python.org/3/library/io.html#io.BufferedIOBase
        https://github.com/python/cpython/blob/1ed61617a4a6632905ad6a0b440cd2cafb8b6414/Lib/_pyio.py#L1461

    """
    def __init__(self, uri):
        StorageObject.__init__(self, uri)
        # Subclasses can use the following attributes
        self._closed = True
        # The following can be set by calling __set_mode(mode)
        # Raw IO always operates in binary mode
        self._mode = None
        self._created = False
        self._readable = False
        self._writable = False
        self._appending = False

    def __str__(self):
        """The URI of the file.
        """
        return self.uri

    def __call__(self, mode='rb'):
        self._set_mode(mode)
        return self

    @property
    def closed(self):
        return self._closed

    @property
    def mode(self):
        return self._mode

    def _set_mode(self, mode):
        """Sets attributes base on the mode.

        See Also: https://docs.python.org/3/library/functions.html#open

        """
        self._mode = mode
        # The following code is modified based on the __init__() of python FileIO class
        if not set(mode) <= set('xrwab+'):
            raise ValueError('Invalid mode: %s' % (mode,))
        if sum(c in 'rwax' for c in mode) != 1 or mode.count('+') > 1:
            raise ValueError('Must have exactly one of create/read/write/append '
                             'mode and at most one plus')

        if 'x' in mode:
            self._created = True
            self._writable = True
        elif 'r' in mode:
            self._readable = True
        elif 'w' in mode:
            self._writable = True
        elif 'a' in mode:
            self._writable = True
            self._appending = True
        if '+' in mode:
            self._readable = True
            self._writable = True

    def _is_same_mode(self, mode):
        if not self.mode:
            return False
        if mode:
            return sorted(self.mode) == sorted(mode)
        return True

    def open(self, mode='r', *args, **kwargs):
        if not self._is_same_mode(mode):
            self._set_mode(mode)
        self._closed = False
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return

    def _check_readable(self):
        """Checks if the file is readable, raise an UnsupportedOperation exception if not.
        """
        if not self.readable():
            raise UnsupportedOperation("File is not opened for read.")

    def _check_writable(self):
        """Checks if the file is writable, raise an UnsupportedOperation exception if not.
        """
        if not self.writable():
            raise UnsupportedOperation("File is not opened for write.")

    def writable(self):
        """Writable if file is writable and not closed.
        """
        return True if self._writable and not self._closed else False

    def readable(self):
        """Returns True if the file exists and readable, otherwise False.
        """
        if self._readable:
            return True
        return False

    def readall(self):
        """Reads all data from the file.

        Returns (bytes): All data in the file as bytes.

        """
        return self.read()

    def readinto(self, b):
        """Reads bytes into a pre-allocated bytes-like object b.

        Returns: An int representing the number of bytes read (0 for EOF), or
            None if the object is set not to block and has no data to read.

        This function is copied from FileIO.readinto()
        """
        # Copied from FileIO.readinto()
        m = memoryview(b).cast('B')
        data = self.read(len(m))
        n = len(data)
        try:
            m[:n] = data
        except Exception as e:
            logger.debug(n)
            logger.debug(len(m))
            raise e
        return n

    @property
    def size(self):
        return None

    @property
    def md5_hex(self):
        return None

    @property
    def updated_time(self):
        """Last updated/modified time of the file as a datetime object.
        """
        raise NotImplementedError()

    def close(self):
        self._closed = True
        raise NotImplementedError("close() is not implemented for %s" % self.__class__.__name__)

    def read(self, size=None):
        """Reads at most "size" bytes.

        Args:
            size: The maximum number of bytes to be returned

        Returns (bytes): At most "size" bytes from the file.
            Returns empty bytes object at EOF.

        """
        self._check_readable()
        raise NotImplementedError()

    def write(self, b):
        """Writes bytes b to file.

        Returns: The number of bytes written into the file.
            None if the write would block.
        """
        self._check_writable()
        raise NotImplementedError()

    def exists(self):
        """Checks if the file exists.
        """
        raise NotImplementedError("exists() is not implemented for %s" % self.__class__.__name__)

    def delete(self):
        raise NotImplementedError()

    def load_from(self, stream):
        """Creates/Loads the file from a stream
        """
        if self.closed:
            with self.open("wb") as f:
                file_size = self.copy_stream(stream, f)
        else:
            file_size = self.copy_stream(stream, self)
        return file_size

    def download(self, to_file_obj):
        raise UnsupportedOperation()

    def upload(self, from_file_obj):
        raise UnsupportedOperation()


class StorageIOSeekable(StorageIOBase, ABC):
    """Base class for seekable Storage
    Seekable storage sub-class should implement:
        seek()
        tell()

    This class has an _offset attribute to help keeping track of the read/write position of the file.
    A sub-class may not use the _offset attribute if the underlying IO keeps track of the position.
    However, if the _offset is used, the read() and write() in the sub-class are responsible to update the _offset.
    Otherwise the _offset will always be 0.

    _seek() provides a simple implementation of seek().
    
    """
    def __init__(self, uri):
        StorageIOBase.__init__(self, uri)
        self._offset = 0

    def seekable(self):
        return True

    def _seek(self, pos, whence=SEEK_SET):
        """Move to new file position.
        Argument offset is a byte count.  Optional argument whence defaults to
        SEEK_SET or 0 (offset from start of file, offset should be >= 0); other values
        are SEEK_CUR or 1 (move relative to current position, positive or negative),
        and SEEK_END or 2 (move relative to end of file, usually negative, although
        many platforms allow seeking beyond the end of a file).
        Note that not all file objects are seekable.
        """
        if not isinstance(pos, int):
            raise TypeError('pos must be an integer.')
        if whence == 0:
            if pos < 0:
                raise ValueError("negative seek position %r" % (pos,))
            self._offset = pos
        elif whence == 1:
            self._offset = max(0, self._offset + pos)
        elif whence == 2:
            self._offset = max(0, self.size + pos)
        else:
            raise ValueError("whence must be 0, 1 or 2.")
        return self._offset

    @property
    def size(self):
        """Returns the size in bytes of the file as an integer.
        """
        raise NotImplementedError()

    def seek(self, pos, whence=SEEK_SET):
        raise NotImplementedError()

    def tell(self):
        raise NotImplementedError()



import os
import shutil
import logging
import datetime
import hashlib
from io import FileIO, SEEK_SET
from .base import StoragePrefixBase, StorageIOSeekable, StorageFolderBase
logger = logging.getLogger(__name__)


class LocalFolder(StorageFolderBase):
    def exists(self):
        logger.debug(self.path)
        return True if os.path.exists(self.path) else False

    @property
    def object_paths(self):
        return [os.path.join(self.path, f) for f in os.listdir(self.path)]

    @property
    def file_paths(self):
        return list(filter(lambda x: os.path.isfile(x), self.object_paths))

    @property
    def folder_paths(self):
        return list(filter(lambda x: os.path.isdir(x), self.object_paths))

    def create(self):
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        return self

    def copy(self, to, contents_only=False):
        """Copies a folder and the files/folders in it.

        Args:
            to (str): The destination path.
            contents_only: Copies only the content of the folder.
                Defaults to False, i.e. a folder (with the same name as this folder)
                will be created at the destination to contain the files.
        """
        if not contents_only:
            to += self.basename
        logger.debug("Copying files from %s to %s" % (self.path, to))
        # Copy the files recursively
        # copytree is not used here as it raises permission denied error in some python versions.
        local_path = LocalFolder(to).path
        # Create the folder if it does not exist
        if not os.path.exists(local_path):
            logger.debug("Creating new folder: %s" % local_path)
            os.makedirs(local_path)
        for file_path in self.file_paths:
            logger.debug("Copying %s" % file_path)
            # Copy the file into the directory
            shutil.copy(file_path, local_path)

        for folder_path in self.folder_paths:
            LocalFolder(folder_path).copy(to)

    def delete(self):
        if os.path.exists(self.path):
            shutil.rmtree(self.path)


class LocalFile(StorageIOSeekable):
    def __init__(self, uri):
        # file_io will be initialized by open()
        self.file_io = None
        StorageIOSeekable.__init__(self, uri)

    # LocalFile supports low level API: fileno() and isatty()
    def fileno(self):
        return self.file_io.fileno()

    def isatty(self):
        return self.file_io.isatty()

    def tell(self):
        return self.file_io.tell()

    def seek(self, pos, whence=SEEK_SET):
        return self.file_io.seek(pos, whence)

    def seekable(self):
        return self.file_io.seekable()

    def read(self, size=None):
        self._check_readable()
        return self.file_io.read(size)

    def write(self, b):
        self._check_writable()
        n = self.file_io.write(b)
        return n

    def truncate(self, size=None):
        return self.file_io.truncate(size)

    def close(self):
        self._closed = True
        if self.file_io:
            self.file_io.close()
        self.file_io = None

    @property
    def size(self):
        """File size in bytes"""
        if self.exists():
            return os.path.getsize(self.path)
        return None

    @property
    def md5_hex(self):
        hash_md5 = hashlib.md5()
        with open(self.path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @property
    def updated_time(self):
        return datetime.datetime.fromtimestamp(os.path.getmtime(self.path))

    def exists(self):
        return True if os.path.exists(self.path) else False

    def delete(self):
        """Deletes the file if it exists.
        """
        if os.path.exists(self.path):
            os.remove(self.path)

    def copy(self, to):
        """Copies the file to another location.
        """
        shutil.copyfile(self.path, LocalFile(to).path)

    def open(self, mode='r', closefd=True, opener=None):
        """
        """
        super().open(mode)
        if self.file_io:
            self.file_io.close()
            self.file_io = None
        if not self.file_io:
            mode = "".join([c for c in self.mode if c in "rw+ax"])
            # Create folder structure if one does not exists
            dir_name = os.path.dirname(self.path)
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)
            self.file_io = FileIO(self.path, mode, closefd=closefd, opener=opener)
        return self


class LocalPrefix(StoragePrefixBase):
    @property
    def uri_list(self):
        uri_list = []
        dir_name = os.path.dirname(self.path)
        parent = LocalFolder(dir_name)
        obj_paths = [p for p in parent.object_paths if p.startswith(self.path)]
        for obj_path in obj_paths:
            if os.path.isfile(obj_path):
                uri_list.append(obj_path)
                continue
            if obj_path == self.path or obj_path in uri_list:
                continue
            if os.path.isdir(obj_path):
                if not obj_path.endswith("/"):
                    obj_path += "/"
                uri_list.extend(LocalPrefix(obj_path).uri_list)
                continue
        return uri_list



"""Contains tests for the web module.
"""
import datetime
import logging
import os
import sys
logger = logging.getLogger(__name__)
try:
    from ..test import AriesTest
    from ..docker import DockerImage, DockerAPI
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.test import AriesTest
    from Aries.docker import DockerImage, DockerAPI


class TestDocker(AriesTest):
    def test_docker_api(self):
        registries = [
            "registry-1.docker.io",
            "gcr.io",
            "quay.io"
        ]
        for registry in registries:
            api = DockerAPI(registry)
            self.assertTrue(api.check_version().startswith("2."))

    def test_docker_image(self):
        image = DockerImage("ubuntu")
        self.assertEqual(image.hostname, "registry-1.docker.io")
        self.assertEqual(image.path, "library/ubuntu")
        self.assertEqual(image.tag, "latest")

        image = DockerImage("ubuntu:16.04")
        self.assertEqual(image.hostname, "registry-1.docker.io")
        self.assertEqual(image.path, "library/ubuntu")
        self.assertEqual(image.tag, "16.04")

        image = DockerImage("us.gcr.io/davelab-gcloud/cancer:0.1-alpha")
        self.assertEqual(image.hostname, "us.gcr.io")
        self.assertEqual(image.path, "davelab-gcloud/cancer")
        self.assertEqual(image.tag, "0.1-alpha")
        self.assertEqual(image.get_size(), 460399799)

        image = DockerImage("quay.io/biocontainers/fastqc:0.11.5--1")
        self.assertEqual(image.hostname, "quay.io")
        self.assertEqual(image.path, "biocontainers/fastqc")
        self.assertEqual(image.tag, "0.11.5--1")
        self.assertEqual(image.get_size(), 148276937)



import logging
import os
import sys
logger = logging.getLogger(__name__)
try:
    from ..test import AriesTest
    from ..oauth import GoogleOAuth
    from ..Google.drive import GoogleSheet
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.test import AriesTest
    from Aries.oauth import GoogleOAuth
    from Aries.Google.drive import GoogleSheet


class TestGoogleSheet(AriesTest):

    google_client_id = None
    google_client_secret = None
    google_refresh_token = None
    google_access_token = None
    google_sheet_id = "1ZTzD1VaLKffRIw8JDhZyLJeUSIwbd7AsJtv8u5XyqrM"

    @classmethod
    def setUpClass(cls):
        cls.google_client_id = os.environ.get("GOOGLE_CLIENT_ID")
        cls.google_client_secret = os.environ.get("GOOGLE_CLIENT_SECRET")
        cls.google_refresh_token = os.environ.get("GOOGLE_REFRESH_TOKEN")

    def setUp(self):
        if not self.google_client_id:
            self.skipTest("GOOGLE_CLIENT_ID not found.")
        if not self.google_client_secret:
            self.skipTest("GOOGLE_CLIENT_SECRET not found.")
        if not self.google_refresh_token:
            self.skipTest("GOOGLE_REFRESH_TOKEN not found.")

    def get_access_token(self):
        if not self.google_access_token:
            self.google_access_token = GoogleOAuth(
                self.google_client_id,
                self.google_client_secret
            ).refresh_access_token(self.google_refresh_token)
            self.assertIsNotNone(self.google_access_token)
        return self.google_access_token

    def test_access_google_sheet(self):
        google_sheet = GoogleSheet(
            file_id=self.google_sheet_id,
            access_token=self.get_access_token()
        )
        data = google_sheet.get_data_grid()
        self.assertGreater(len(data), 1)
        self.assertEqual(data[0][0], "A1")

        row = google_sheet.get_row_data(2, sheet_name="Sheet1")
        self.assertEqual(row[2], "C3")
        self.assertEqual(len(row), 3)



"""Contains tests for the storage package.
"""
import datetime
import logging
import os
import sys
import time
import traceback
logger = logging.getLogger(__name__)


try:
    from ..test import AriesTest
    from ..storage import StoragePrefix, StorageFolder, StorageFile
    from ..storage import gs
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.storage import gs
    from Aries.test import AriesTest
    from Aries.storage import StoragePrefix, StorageFile, StorageFolder


class TempFolder:
    def __init__(self, folder_uri):
        self.folder_uri = folder_uri

    def __enter__(self):
        folder = StorageFolder(self.folder_uri)
        if not folder.exists():
            folder.create()
        return folder

    def __exit__(self, exc_type, exc_val, exc_tb):
        folder = StorageFolder(self.folder_uri)
        if folder.exists():
            folder.delete()


class TestStorage(AriesTest):
    SCHEME = "file"
    HOST = ""
    TEST_ROOT_PATH = os.path.join(os.path.dirname(__file__), "fixtures", "test_folder")
    TEST_ROOT = "%s://%s%s" % (SCHEME, HOST, TEST_ROOT_PATH)
    test_folder = StorageFolder(TEST_ROOT)

    @classmethod
    def create_folder(cls, relative_path):
        abs_path = os.path.join(cls.TEST_ROOT, relative_path)
        StorageFolder(abs_path).create()

    @classmethod
    def create_file(cls, relative_path, content):
        """Creates a file relative to the test root
        """
        abs_path = os.path.join(cls.TEST_ROOT, relative_path)
        with StorageFile.init(abs_path, "w") as f:
            f.write(content)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if not cls.test_folder.exists():
            cls.test_folder.create()
        cls.test_folder.empty()
        cls.create_folder("test_folder_0")
        cls.create_folder("test_folder_1")
        cls.create_file("file_in_test_folder", "")
        cls.create_file("test_folder_0/empty_file", "")
        cls.create_file("test_folder_0/abc.txt", "abc\ncba\n")
        cls.create_file("test_folder_1/.hidden_file", "")

    @classmethod
    def tearDownClass(cls):
        StorageFolder(cls.TEST_ROOT).empty()
        StorageFolder(cls.TEST_ROOT).delete()
        super().tearDownClass()

    def setUp(self):
        super().setUp()

    def test_prefix(self):
        storage_prefix = StoragePrefix(os.path.join(self.TEST_ROOT, "test_folder"))
        self.assertEqual(len(storage_prefix.objects), 3)

    def test_storage_folder(self):
        # Scheme
        self.assertEqual(self.test_folder.scheme, self.SCHEME)
        # Hostname
        if self.test_folder.hostname or self.HOST:
            self.assertEqual(self.test_folder.hostname, self.HOST)
        # Folder URI will always end with "/"
        self.assertTrue(str(self.test_folder).endswith("/"))
        # Path
        self.assertEqual(self.test_folder.path.rstrip("/"), self.TEST_ROOT_PATH.rstrip("/"))
        # Name
        self.assertEqual(self.test_folder.basename, os.path.basename(self.TEST_ROOT_PATH))
        self.assertEqual(self.test_folder.name, os.path.basename(self.TEST_ROOT_PATH))
        # Folder attributes
        self.assertGreater(len(self.test_folder.get_folder_attributes()), 1)
        # File attributes
        self.assertIn("file_in_test_folder", self.test_folder.get_file_attributes("name"))

    def test_get_folder_names(self):
        folder = StorageFolder(self.TEST_ROOT)
        folder_names = folder.folder_names
        # There are two sub-folders
        self.assertEqual(len(folder_names), 2, 'Folders: %s' % folder_names)
        self.assertIn("test_folder_0", folder_names)
        self.assertIn("test_folder_1", folder_names)
        # "/" should not be found in folder names
        self.assertNotIn("/", str(folder_names))

    def test_get_paths(self):
        sub_folders = self.test_folder.folder_paths
        # There are two sub-folders
        self.assertEqual(len(sub_folders), 2, "Folders: %s" % sub_folders)
        # TODO: local paths not having scheme?
        if self.SCHEME == "file":
            self.assertIn(os.path.join(self.TEST_ROOT_PATH, "test_folder_0"), sub_folders)
            self.assertIn(os.path.join(self.TEST_ROOT_PATH, "test_folder_1"), sub_folders)
        else:
            # TODO: folders ends with "/"??
            self.assertIn(os.path.join(self.TEST_ROOT, "test_folder_0/"), sub_folders)
            self.assertIn(os.path.join(self.TEST_ROOT, "test_folder_1/"), sub_folders)

    def test_get_folder_and_file(self):
        # Try to get a folder that does not exist. Should return None
        self.assertIsNone(self.test_folder.get_folder("not_exist"))

        # Get a sub folder
        sub_folder = self.test_folder.get_folder("test_folder_1")
        self.assertTrue(
            isinstance(sub_folder, StorageFolder),
            "Failed to get sub folder. Received %s: %s" % (sub_folder.__class__, sub_folder))

        # Get the filename of hidden_file
        filenames = sub_folder.file_names
        self.assertEqual(len(filenames), 1)
        self.assertEqual(filenames[0], ".hidden_file")

        # Get a file that does not exist
        f = sub_folder.get_file("not_exist")
        self.assertIsNone(f)

        # Get the hidden file
        f = sub_folder.get_file(".hidden_file")
        self.assertTrue(f.exists(), "Failed to get the file.")
        self.assertEqual(f.basename, ".hidden_file")
        self.assertTrue(
            isinstance(f, StorageFile),
            "Failed to get storage file from a folder. Received %s: %s" % (f.__class__, f))

    def test_text_read(self):
        with StorageFile.init(os.path.join(self.TEST_ROOT, "file_in_test_folder")) as f:
            self.assertEqual(f.size, 0)
            self.assertEqual(f.tell(), 0)
            self.assertEqual(f.seek(0, 2), 0)
            self.assertEqual(len(f.read()), 0)

    def test_text_read_write(self):
        # Write a new file
        temp_file_path = os.path.join(self.TEST_ROOT, "temp_file.txt")
        with StorageFile.init(temp_file_path, 'w+') as f:
            self.assertTrue(f.writable())
            self.assertEqual(f.tell(), 0)
            self.assertEqual(f.write("abc"), 3)
            self.assertEqual(f.tell(), 3)
            f.seek(0)
            self.assertEqual(f.read(), "abc")
            # TODO: File may not exist on the cloud until it is closed.
            # self.assertTrue(f.exists())
        f.delete()

    def test_binary_read_write(self):
        # File does not exist, a new one will be created
        file_uri = os.path.join(self.TEST_ROOT, "test.txt")
        storage_file = StorageFile(file_uri).open("wb")
        self.assertEqual(storage_file.scheme, self.SCHEME)
        self.assertTrue(storage_file.seekable())
        self.assertFalse(storage_file.readable())
        self.assertEqual(storage_file.write(b"abc"), 3)
        self.assertEqual(storage_file.tell(), 3)
        self.assertEqual(storage_file.write(b"def"), 3)
        self.assertEqual(storage_file.tell(), 6)
        storage_file.close()
        self.assertTrue(storage_file.exists())
        storage_file.open('rb')
        self.assertEqual(storage_file.read(), b"abcdef")
        storage_file.close()
        storage_file.delete()
        self.assertFalse(storage_file.exists())

    def test_copy_folder(self):
        # Source folder to be copied
        src_folder_uri = os.path.join(self.TEST_ROOT, "test_folder_0")
        sub_folder = StorageFolder(src_folder_uri)
        # Source folder should exist
        self.assertTrue(sub_folder.exists())

        # Destination folder
        dst_folder_uri = os.path.join(self.TEST_ROOT, "new_folder", "test_folder_0")

        dst_parent = os.path.join(self.TEST_ROOT, "new_folder")
        with TempFolder(dst_parent):
            # Destination folder should not exist
            dst_folder = StorageFolder(dst_folder_uri)
            if dst_folder.exists():
                dst_folder.delete()
            self.assertFalse(dst_folder.exists())

            # Copy the folder
            if not dst_parent.endswith("/"):
                dst_parent += "/"
            logger.debug("Copying from %s into %s" % (sub_folder.uri, dst_parent))
            sub_folder.copy(dst_parent)

            # Destination folder should now exist and contain an empty file
            self.assertTrue(dst_folder.exists())
            file_paths = [f.uri for f in dst_folder.files]
            self.assertEqual(len(file_paths), 2)
            self.assertIn(os.path.join(dst_folder_uri, "empty_file"), file_paths)
            self.assertIn(os.path.join(dst_folder_uri, "abc.txt"), file_paths)

    def test_create_copy_and_delete_file(self):
        new_folder_uri = os.path.join(self.TEST_ROOT, "new_folder")
        with TempFolder(new_folder_uri) as folder:
            self.assertTrue(folder.is_empty())

            # Create a sub folder inside the new folder
            sub_folder_uri = os.path.join(new_folder_uri, "sub_folder")
            logger.debug(sub_folder_uri)
            sub_folder = StorageFolder(sub_folder_uri).create()
            self.assertTrue(sub_folder.exists())

            # Copy an empty file
            src_file_path = os.path.join(self.TEST_ROOT, "test_folder_0", "empty_file")
            dst_file_path = os.path.join(new_folder_uri, "copied_file")
            f = StorageFile(src_file_path)
            logger.debug(f.exists())
            time.sleep(2)
            f.copy(dst_file_path)
            self.assertTrue(StorageFile(dst_file_path).exists())

            # Copy a file with content and replace the empty file
            src_file_path = os.path.join(self.TEST_ROOT, "test_folder_0", "abc.txt")
            dst_file_path = os.path.join(new_folder_uri, "copied_file")
            f = StorageFile(src_file_path)
            f.copy(dst_file_path)
            dst_file = StorageFile(dst_file_path)
            self.assertTrue(dst_file.exists())
            # Use the shortcut to read file, the content will be binary.
            self.assertEqual(dst_file.read(), b"abc\ncba\n")

            # Empty the folder. This should delete file and sub folder only
            folder.empty()
            self.assertTrue(folder.exists())
            self.assertTrue(folder.is_empty())
            self.assertFalse(sub_folder.exists())
            self.assertFalse(dst_file.exists())


class TestStorageGCP(TestStorage):
    SCHEME = "gs"
    HOST = "aries_test"
    TEST_ROOT_PATH = "/storage_test"
    TEST_ROOT = "%s://%s%s" % (SCHEME, HOST, TEST_ROOT_PATH)
    test_folder = StorageFolder(TEST_ROOT)

    @classmethod
    def setUpClass(cls):
        cls.CREDENTIALS = False
        # Google credentials are required for setting up the class.
        gs.setup_credentials("GOOGLE_CREDENTIALS", os.path.join(os.path.dirname(__file__), "gcp.json"))
        try:
            super().setUpClass()
            cls.CREDENTIALS = True
        except Exception as ex:
            print("%s: %s" % (type(ex), str(ex)))
            traceback.print_exc()

    def setUp(self):
        # Skip test if GCP_ACCESS is not True.
        if not self.CREDENTIALS:
            self.skipTest("GCP Credentials not found.")
        super().setUp()
        time.sleep(0.5)


class TestStorageAWS(TestStorage):
    SCHEME = "s3"
    HOST = "davelab-test"
    TEST_ROOT_PATH = "/storage_test"
    TEST_ROOT = "%s://%s%s" % (SCHEME, HOST, TEST_ROOT_PATH)
    test_folder = StorageFolder(TEST_ROOT)

    @classmethod
    def setUpClass(cls):
        cls.CREDENTIALS = False
        # AWS credentials are loaded from environment variable directly.
        try:
            super().setUpClass()
            cls.CREDENTIALS = True
        except Exception as ex:
            print("%s: %s" % (type(ex), str(ex)))
            traceback.print_exc()
            raise ex

    def setUp(self):
        # Skip test if self.CREDENTIALS is not True.
        if not self.CREDENTIALS:
            self.skipTest("Credentials for %s not found." % self.__class__.__name__)
        super().setUp()
        time.sleep(0.5)



class DictList(list):
    """Represent a list of dictionaries
    """
    def unique_keys(self):
        """Gets the unique keys (union) of the dictionaries
        
        Returns:
            list: A list of keys from the dictionaries
        """
        keys = set()
        for d in self:
            for key in d.keys():
                keys.add(key)
        return keys

    def sort_by_value(self, key, reverse=False):
        """Sort the dictionaries by the value of a particular keys.
        
        Args:
            key: A key in the dictionaries.
            reverse (bool, optional): Reverse sorting. Defaults to False.
        
        Returns:
            list: A list of sorted dictionaries.
        """
        return sorted(self, key=lambda i: i.get(key), reverse=reverse)


class NestedList(list):
    """Represents a list of lists.

    """
    def __init__(self, *args):
        """Initializes a nested list.

        Args:
            *args: A list of lists.

        """
        super().__init__(args)

    def sort_elements(self, **kwargs):
        """Sorts the elements in each list based on the values of the first list in the nested list.
        When the values in the first list are the same, the values in the second list will be used.

        Args:
            **kwargs: keyword arguments for the built-in sorted() function.

        Returns: A sorted NestedList object.

        """
        if len(self) == 0:
            return []
        return NestedList(*[list(t) for t in zip(*sorted(zip(*self), **kwargs))])

    def flatten(self):
        """Flatten the lists into a single list
        """
        return [element for sub_list in self for element in sub_list]


def sort_lists(order_list, label_list, reverse=False):
    """Sort a two lists at the same time.
    The order will be determined by the order_list.

    Example:
        order_list = [2, 1, 3]
        label_list = ['A', 'B', 'C']
        After sorting,
        order_list = [1, 2, 3]
        label_list = ['B', 'A', 'C']
    
    Args:
        order_list (list): The list to be sorted.
        label_list (list): This list will be rearrange based on the order of order_list.
        reverse (bool, optional): Reverse sorting. Defaults to False.
    
    Returns: A 2-tuple of sorted (order_list, label_list).
    """
    # Do nothing if the order_list if None or empty.
    if not order_list:
        return order_list, label_list
    sorted_lists = NestedList(order_list, label_list).sort_elements(reverse=reverse)
    return sorted_lists[0], sorted_lists[1]



import logging
from ..web import WebAPI
logger = logging.getLogger(__name__)


class GoogleDriveFile:
    def __init__(self, file_id, access_token):
        self.file_id = file_id
        self._metadata = None
        self.api = WebAPI("https://www.googleapis.com/")
        self.api.add_header(Authorization="Bearer %s" % access_token)

    @property
    def metadata(self):
        """
        See Also: https://developers.google.com/drive/api/v3/fields-parameter
        """
        if not self._metadata:
            url = "https://www.googleapis.com/drive/v3/files/%s?fields=*" % self.file_id
            self._metadata = self.api.get_json(url)
        return self._metadata

    def get_meta(self, fields):
        """
        See Also: https://developers.google.com/drive/api/v3/fields-parameter
        """
        if isinstance(fields, list):
            fields = "&".join(["fields=%s" % f for f in fields])
            url = "https://www.googleapis.com/drive/v3/files/%s?%s" % (self.file_id, fields)
        else:
            url = "https://www.googleapis.com/drive/v3/files/%s?fields=%s" % (self.file_id, fields)
        return self.api.get_json(url)

    @property
    def revisions(self):
        return self.api.get_json("https://www.googleapis.com/drive/v3/files/%s/revisions" % self.file_id)


class GoogleSheet(GoogleDriveFile):
    """Represents a Google Sheet file
    """
    @property
    def sheets(self):
        """Gets a list of sheets from the API response (in terms of dictionaries).

        Returns:
            list: A list of dictionaries.
            Each dictionary contains sheet information as defined at
            https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#Sheet
            This list is the value of the "sheets" key in the response of the response of
            GET https://sheets.googleapis.com/v4/spreadsheets/{spreadsheetId}

        """
        return self.get().get("sheets")

    def get(self, **kwargs):
        """Returns properties of the spreadsheet along with properties and merges of sheets in the spreadsheet.
        By default, data within grids will not be returned. You can include grid data one of two ways:
            1. Specify a field mask listing your desired fields using the fields parameter.
            2. Use key word argument "includeGridData=True".
        If a field mask is set, the includeGridData parameter is ignored
        For large spreadsheets, it is recommended to retrieve only the specific fields of the spreadsheet that you want.

        To retrieve only subsets of the spreadsheet, use the ranges parameter. Multiple ranges can be specified.
        Limiting the range will return only the portions of the spreadsheet that intersect the requested ranges.
        Ranges are specified using A1 notation.

        Args:
            **kwargs: ranges, includeGridData, fields

        Returns: A dictionary containing the information of the spreadsheet.

        See Also:
            https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get
            https://developers.google.com/sheets/api/guides/concepts#partial_responses
        """
        api_url = "https://sheets.googleapis.com/v4/spreadsheets/%s" % self.file_id
        return self.api.get_json(api_url, **kwargs)

    def values(self, data_range, **kwargs):
        """Gets the values of a specific range.

        Args:
            data_range: A1 notation of a range in a sheet.
                e.g. "SheetName!A1:B5"

        Returns: A dictionary containing the following keys:
            majorDimension:
            range: The actual range in the sheet containing data.
            values: Two 2-D list containing the values.
        """
        return self.api.get_json(
            "https://sheets.googleapis.com/v4/spreadsheets/%s/values/%s" % (self.file_id, data_range),
            **kwargs
        )

    def get_data_grid(self, sheet_index=0, value_type="formattedValue"):
        """Gets the values of a sheet

        Args:
            sheet_index: 0-based index of the sheet. Defaults to 0, the first sheet.
            value_type: The type of the values. Defaults to "formattedValue".
                This is a key of the CellData, as defined in
                https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData
                Other string types include hyperlink and notes.

        Returns: A 2-D list of values. The type of the values depends on the value_type parameter.

        """
        sheets = self.get(includeGridData=True).get("sheets")
        if not sheets:
            return None

        data = sheets[sheet_index].get("data")
        if not data:
            return None

        rows = data[0].get("rowData", [])
        grid = []
        for row in rows:
            values = row.get("values", [])
            row_values = [v.get(value_type) for v in values]
            grid.append(row_values)
        return grid

    def get_row_data(self, row_number, sheet_name, from_col=None):
        """Gets the data values of a row from a sheet as a list

        Args:
            row_number (int): The 1-based row number.
            sheet_name (str): The name of the sheet.
            from_col (int): Gets the data starting from a certain column.
                This can be used to exclude the values header columns.
                All values of the rows will be returned if from_col is None, 0 or evaluated as False.

        Returns: A list of values.
            The value of an empty cell will be an empty string.


        """
        values = self.values("%s!%s:%s" % (sheet_name, row_number, row_number)).get("values")
        if not values:
            return []
        values = values[0]
        if from_col:
            values = values[from_col:] if from_col < len(values) else []
        return values

    def get_column_data(self, col_name, sheet_name, from_row=None):
        """Gets the data values of a column from a sheet as a list

        Args:
            sheet_name (str): The name of the sheet
            col_name (str): The column as letter string, e.g. "A" or "AK".
            from_row (int): Gets the data starting from a certain row.
                This can be used to exclude the values header rows.
                All values of the column will be returned if from_row is None, 0 or evaluated as False.

        Returns: A list of values.
            The value of an empty cell will be an empty string.


        """
        values = self.values("%s!%s:%s" % (sheet_name, col_name, col_name)).get("values")
        if not values:
            return []
        if from_row:
            values = values[from_row:]
        values = [v[0] if len(v) > 0 else "" for v in values]
        return values

    def append(self, data_range, rows):
        """Appends rows of values to the sheet after a "table" in the data_range.

        Args:
            data_range: The A1 notation of a range to search for a logical table of data.
                Values are appended after the last row of the table.
                For more details about how the table is detected, see:
                https://developers.google.com/sheets/api/guides/values#appending_values
            rows: A 2-D list containing the values to be appended.

        Returns:

        See Also:
            https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/append

        """
        url = "https://sheets.googleapis.com/v4/spreadsheets/%s/values/%s:append?" \
              "valueInputOption=USER_ENTERED&insertDataOption=INSERT_ROWS" % (self.file_id, data_range)
        data = {
            "range": data_range,
            "majorDimension": "ROWS",
            "values": rows
        }
        self.api.post_json(url, data)



"""Contains AriesTest class, a customized class to be used in place of unittest.TestCase.
"""
import logging
import time
import os
import sys
from collections import OrderedDict
from unittest import TestCase
from .outputs import StreamHandler, PackageLogFilter


class AriesTest(TestCase):
    """Customized TestCase:
        1. A stream handler is added to the root logger to output debug messages.
        2. Test outputs are buffered. Outputs for passing tests are discard.
            Only outputs for failed tests will be displayed.

        By default, the steam handler will be added to the root logger.
        This can be changed by override the "logger_names" class attribute 
            to a list of logger names, e.g. logger_names = ["tests"].

    Usage:
        This class can be used to replace unittest.TestCase.
        No other changes is needed.

    Remarks:
        Make sure super().setUpClass() is called when overriding this method in sub-classes.
    
    """
    # Stores the time for each test.
    # key: name of the test method.
    # value: time elapsed during the test.
    times = OrderedDict()

    # Override logger_names in subclasses to avoid enabling debug logging for root logger.
    logger_names = [""]

    decorated = False

    @classmethod
    def __decorate_test_case(cls, func):
        """A decorator for test case.
        1. Adds additional logging handler to test case so that logging will be streamed to stdout.
        2. Set logger level to DEBUG

        Args:
            func: test case function.

        Returns: A function.

        """
        def test_case_with_logging(*args, **kwargs):
            """Wraps the test cases and add a stream handler for logging
            """
            # At this point, unittest already replaced the sys.stdout
            # The sys.stdout passing into the StreamHandler is NOT the original system standard output
            # The default value (sys.stdout) for StreamHanlder() will send the logs to the system standard output
            stream_handler = StreamHandler(sys.stdout)
            package_filter = PackageLogFilter(packages=[os.path.basename(os.path.abspath(os.curdir))])
            stream_handler.addFilter(package_filter)
            loggers = dict()
            for name in cls.logger_names:
                logger = logging.getLogger(name)
                level = logger.getEffectiveLevel()
                loggers[logger] = level
                logger.setLevel(logging.DEBUG)
                logger.addHandler(stream_handler)
            try:
                # Run the function(test)
                results = func(*args, **kwargs)
            finally:
                for logger, level in loggers.items():
                    logger.removeHandler(stream_handler)
                    logger.setLevel(level)
            return results

        return test_case_with_logging

    @classmethod
    def setUpClass(cls):
        """Setup the test class.
        Each test case will be decorated so that the logger outputs are streamed to stdout.

        """
        super().setUpClass()
        if not cls.decorated:
            # Find all test cases
            attrs = dir(cls)
            for attr in attrs:
                if attr.startswith("test") and callable(getattr(cls, attr)):
                    test_case = getattr(cls, attr)
                    setattr(cls, attr, cls.__decorate_test_case(test_case))

            cls.decorated = True

    def run(self, result=None):
        """Runs the test case with standard output buffered.
        Output during a passing test is discarded.
        Output is echoed normally on test fail or error and is added to the failure messages.
        This is the same as running the Python unittest command line with "-b" option.
        
        See also:
        https://docs.python.org/3/library/unittest.html#command-line-options
        https://docs.python.org/3/library/unittest.html#unittest.TestResult.buffer

        Args:
            result:

        Returns:

        """
        if result is None:
            result = self.defaultTestResult()
        result.buffer = True
        start = time.time()
        returns = super().run(result)
        end = time.time()
        elapsed = end - start
        self.times[self._testMethodName] = elapsed
        return returns

